<?php

declare(strict_types=1);

namespace Drupal\group_context_path_prefix;

use Drupal\Core\Database\Connection;
use Drupal\Core\Database\Query\SelectInterface;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityRepositoryInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\Sql\DefaultTableMapping;
use Drupal\Core\Entity\Sql\SqlEntityStorageInterface;
use Drupal\Core\Field\BaseFieldDefinition;
use Drupal\Core\Render\BubbleableMetadata;
use Drupal\Core\Url;
use Drupal\group\Entity\Group;
use Drupal\group\Entity\GroupInterface;
use Drupal\group\Entity\GroupRelationship;
use Drupal\group\Entity\GroupRelationshipInterface;
use Drupal\group_context_path_prefix\Entity\GroupPathPrefix;
use Drupal\group_context_path_prefix\Trait\PathPrefixMatcherTrait;

/**
 * Loads a group by various strategies.
 */
final class GroupPathPrefixRepository implements GroupPathPrefixRepositoryInterface {

  use PathPrefixMatcherTrait;

  public function __construct(
    private readonly Connection $database,
    private readonly EntityTypeManagerInterface $entityTypeManager,
    private readonly EntityFieldManagerInterface $entityFieldManager,
    private readonly EntityRepositoryInterface $entityRepository,
  ) {}

  /**
   * {@inheritdoc}
   */
  public function getGroupByRequestPathPrefix(string $request_path): ?GroupInterface {
    // Normalize trailing slash vs. no slash.
    $request_path = \rtrim($request_path, '/');
    // Build and execute a database query to find a group that matches the
    // request path.
    $results = $this->buildQuery($request_path)->execute();
    if (!\is_null($results)) {
      $matches = $results->fetchAllKeyed();
      // Of all the prefixes which appear at the start of the request path, find
      // the best match.
      $group_id = self::matchPathPrefix($request_path, $matches);
      if ($group_id === FALSE) {
        return NULL;
      }
      return Group::load($group_id);
    }
    return NULL;
  }

  /**
   * {@inheritdoc}
   */
  public function getGroupByInternalPath(string $path, ?BubbleableMetadata $bubbleable_metadata): ?GroupInterface {
    if (\is_null($bubbleable_metadata)) {
      $bubbleable_metadata = new BubbleableMetadata();
    }
    $url = Url::fromUri('internal:' . $path);
    if (!$url->isRouted()) {
      $bubbleable_metadata->addCacheContexts(['url']);
      return NULL;
    }
    $bubbleable_metadata->addCacheContexts(['route.name']);
    $route_parts = \explode('.', $url->getRouteName());
    if (\count($route_parts) !== 3) {
      return NULL;
    }
    // Is this an entity route? E.g., "entity.node.canonical".
    // @todo figure out what to do about Views pages.
    if ($route_parts[0] !== 'entity' || $route_parts[2] !== 'canonical') {
      return NULL;
    }
    $entity_type_id = $route_parts[1];
    $entity_id = $url->getRouteParameters()[$entity_type_id] ?? NULL;
    $bubbleable_metadata->addCacheContexts(['route']);
    if (empty($entity_id)) {
      return NULL;
    }
    // @todo should this use getActive() instead of getCanonical()?
    $entity = $this->entityRepository->getCanonical($entity_type_id, $entity_id);
    if (!$entity instanceof EntityInterface) {
      return NULL;
    }
    $bubbleable_metadata->addCacheableDependency($entity);
    if ($entity instanceof GroupInterface) {
      return $entity;
    }
    $relationships = GroupRelationship::loadByEntity($entity);
    // @todo prefer the current group if the entity belong to multiple groups.
    $group_relationship = \array_shift($relationships);
    if (!$group_relationship instanceof GroupRelationshipInterface) {
      return NULL;
    }
    return $group_relationship->getGroup();
  }

  /**
   * Builds a database query to find group IDs with a matching path prefix.
   *
   * Does not use an entity query because this query executes in the critical
   * path, which would require loading all the selected group IDs in order to
   * get the path field value.
   *
   * @param string $request_path
   *   The incoming request path.
   *
   * @return \Drupal\Core\Database\Query\SelectInterface
   *   A query returning the group ID and prefix for all group prefixes whose
   *   string value is at the beginning of the request path string.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  private function buildQuery(string $request_path): SelectInterface {
    $entity_type = $this->entityTypeManager->getDefinition('group');
    \assert($entity_type instanceof EntityTypeInterface);
    // Which table and column has the group path prefix in it?
    [
      'entity_data_table' => $field_data_table,
      'value_column_name' => $path_prefix_column,
    ] = $this->getPathPrefixTableAndColumn();
    // Which column stores the group ID?
    $id = $entity_type->getKey('id');
    // Build the base query for group IDs and and path prefixes.
    $query = $this->database
      ->select($field_data_table, 'gfd')
      ->fields('gfd', [$id, $path_prefix_column]);
    // Only return rows where the group's path prefix is at the beginning of
    // the request path, i.e., where the index of the path prefix string is at
    // position 1 in the request path. Note, this is not sufficient to find a
    // match on its own. For the reason described by matchPathPrefix().
    /* @link \Drupal\group_context_path_prefix\Trait\PathPrefixMatcherTrait::matchPathPrefix() */
    /* @see https://dev.mysql.com/doc/refman/8.0/en/string-functions.html#function_instr. */
    $query_field = "gfd.{$path_prefix_column}";
    $query->where("INSTR(:request_path, {$query_field}) = 1", [
      ':request_path' => $request_path,
    ]);
    return $query;
  }

  /**
   * Determines the names of the group field data table and path prefix column.
   *
   * @return array
   *   A keyed array whose "entity_data_table" key's value is the group field
   *   data table name and whose "value_column_name" is the column name for the
   *   group's path prefix field value property.
   */
  private function getPathPrefixTableAndColumn(): array {
    $base_field_definitions = $this->entityFieldManager->getBaseFieldDefinitions('group');
    $path_prefix_field_definition = $base_field_definitions[GroupPathPrefix::FIELD_NAME];
    \assert($path_prefix_field_definition instanceof BaseFieldDefinition);
    $storage_handler = $this->entityTypeManager->getStorage('group');
    \assert($storage_handler instanceof SqlEntityStorageInterface);
    $table_mapping = $storage_handler->getTableMapping();
    \assert($table_mapping instanceof DefaultTableMapping);
    return [
      'entity_data_table' => $table_mapping->getDataTable(),
      'value_column_name' => $table_mapping->getFieldColumnName($path_prefix_field_definition, 'value'),
    ];
  }

}
