<?php

declare(strict_types=1);

namespace Drupal\group_context_path_prefix\PathProcessor;

use Drupal\Core\PathProcessor\InboundPathProcessorInterface;
use Drupal\Core\PathProcessor\OutboundPathProcessorInterface;
use Drupal\Core\Render\BubbleableMetadata;
use Drupal\group\Entity\GroupInterface;
use Drupal\group_context_path_prefix\Entity\GroupPathPrefix;
use Drupal\group_context_path_prefix\GroupPathPrefixRepositoryInterface;
use Symfony\Component\HttpFoundation\Request;

/**
 * Processes paths using group path prefixes.
 */
final class GroupPathPrefixProcessor implements InboundPathProcessorInterface, OutboundPathProcessorInterface {

  /**
   * Constructs a new GroupPathPrefixProcessor object.
   *
   * @param \Drupal\group_context_path_prefix\GroupPathPrefixRepositoryInterface $prefixRepository
   *   The "group_context_path_prefix.repository" service.
   */
  public function __construct(
    private readonly GroupPathPrefixRepositoryInterface $prefixRepository,
  ) {}

  /**
   * {@inheritdoc}
   */
  public function processInbound($path, Request $request): string {
    $prefix = $this->getRequestPathPrefix($request);
    if (!\is_string($prefix)) {
      return $path;
    }
    if (!\str_starts_with($path, $prefix)) {
      return $path;
    }
    // Remove the prefix from the path and ensure that the processed path always
    // begins with a forward slash in the case where the prefix exactly matches
    // the inbound path.
    return '/' . \ltrim(\substr($path, \strlen($prefix)), '/');
  }

  /**
   * {@inheritdoc}
   */
  public function processOutbound($path, &$options = [], ?Request $request = NULL, ?BubbleableMetadata $bubbleable_metadata = NULL): string {
    // @todo determine if the language context is necessary.
    /* $langcode = isset($options['language']) ? $options['language']->getId() : NULL; */
    /* $prefix = $this->getGroupPathPrefix($path, $langcode, $bubbleable_metadata); */
    if ($path == "/") {
      return $path;
    }
    $prefix = $this->getGroupPathPrefix($path, $bubbleable_metadata);
    if (!\is_string($prefix)) {
      if (!$request instanceof Request) {
        return $path;
      }
      $prefix = $this->getRequestPathPrefix($request);
      if (!\is_string($prefix)) {
        return $path;
      }
    }
    return $prefix . \rtrim($path, '/');
  }

  /**
   * Gets the group path prefix from the request URL, if it contains one.
   *
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The request, or NULL.
   *
   * @return string|null
   *   The path prefix or NULL if the request URL does not contain a group path
   *   prefix.
   */
  private function getRequestPathPrefix(Request $request): ?string {
    $group = $this->prefixRepository->getGroupByRequestPathPrefix($request->getPathInfo());
    if (!$group instanceof GroupInterface) {
      return NULL;
    }
    return GroupPathPrefix::get($group);
  }

  /**
   * Gets a group path prefix if the path identifies an entity in a group.
   *
   * @param string $path
   *   The path to check.
   * @param \Drupal\Core\Render\BubbleableMetadata|null $bubbleable_metadata
   *   (optional) A BubbleableMetadata object.
   *
   * @return string|null
   *   The path prefix or NULL if the path does not have one.
   */
  private function getGroupPathPrefix(string $path, ?BubbleableMetadata $bubbleable_metadata): ?string {
    $group = $this->prefixRepository->getGroupByInternalPath($path, $bubbleable_metadata);
    if (!$group instanceof GroupInterface) {
      return NULL;
    }
    if (!\is_null($bubbleable_metadata)) {
      $bubbleable_metadata->addCacheTags($group->getCacheTags());
    }
    return GroupPathPrefix::get($group);
  }

}
