<?php

declare(strict_types=1);

namespace Drupal\Tests\group_context_path_prefix\Functional;

use Drupal\Core\Url;
use Drupal\group\Entity\GroupInterface;
use Drupal\group\PermissionScopeInterface;
use Drupal\group\Plugin\Group\RelationHandler\PermissionProviderInterface;
use Drupal\group_context_path_prefix\Entity\GroupPathPrefix;
use Drupal\node\NodeInterface;
use Drupal\Tests\group\Functional\GroupBrowserTestBase;
use Drupal\user\RoleInterface;
use Drupal\user\UserInterface;

/**
 * High-level integration test of the module.
 *
 * Tests the various functions of the module when integrated together and with
 * features of the group, group_sites, and gnode module.
 *
 * @group group_context_path_prefix
 */
final class GroupContextPathPrefixIntegrationTest extends GroupBrowserTestBase {

  /**
   * {@inheritdoc}
   */
  protected static $modules = [
    'gnode',
    'group',
    'group_context_path_prefix',
    'group_sites',
  ];

  /**
   * {@inheritdoc}
   */
  protected $defaultTheme = 'stark';

  /**
   * The test user.
   */
  protected readonly UserInterface $testUser;

  /**
   * A test group.
   */
  protected readonly GroupInterface $groupA;

  /**
   * Another test group.
   */
  protected readonly GroupInterface $groupB;

  /**
   * A test node.
   */
  protected readonly NodeInterface $nodeA;

  /**
   * Another test node.
   */
  protected readonly NodeInterface $nodeB;

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();
    // Create a test group type.
    $this->groupType = $this->createGroupType(['creator_membership' => FALSE]);
    // Create a group admin.
    $this->testUser = $this->drupalCreateUser(['access content overview']);
    // Create test groups.
    $this->groupA = $this->createGroup([
      'type' => $this->groupType->id(),
      'group_context_path_prefix' => '/' . $this->randomMachineName(),
    ]);
    $this->groupB = $this->createGroup([
      'type' => $this->groupType->id(),
      'group_context_path_prefix' => '/' . $this->randomMachineName(),
    ]);
    // Create test content type.
    $node_type = $this->drupalCreateContentType([
      'type' => 'article',
      'name' => 'Article',
    ]);
    // Enable the gnode content plugin.
    $group_plugin_id = "group_node:{$node_type->id()}";
    $this->entityTypeManager->getStorage('group_relationship_type')
      ->createFromPlugin($this->groupType, $group_plugin_id)->save();
    // Set up a group role.
    $node_permission_provider = $this->container->get('group_relation_type.manager')
      ->getPermissionProvider('group_node:article');
    \assert($node_permission_provider instanceof PermissionProviderInterface);
    $this->createGroupRole([
      'group_type' => $this->groupType->id(),
      'scope' => PermissionScopeInterface::OUTSIDER_ID,
      'permissions' => [
        'view group',
        $node_permission_provider->getPermission('view', 'entity'),
      ],
      'global_role' => RoleInterface::AUTHENTICATED_ID,
    ]);
    // Create two test nodes.
    $this->nodeA = $this->drupalCreateNode([
      'type' => $node_type->id(),
      'uid' => $this->testUser->id(),
      'status' => 1,
    ]);
    $this->nodeB = $this->drupalCreateNode([
      'type' => $node_type->id(),
      'uid' => $this->testUser->id(),
      'status' => 1,
    ]);
    // Place each node in its respective group.
    $this->groupA->addRelationship($this->nodeA, $group_plugin_id);
    $this->groupB->addRelationship($this->nodeB, $group_plugin_id);
    // Enable the group_context_path_prefix.path_prefix_context service with the
    // group_sites module.
    $group_sites_settings = $this->config('group_sites.settings');
    $group_sites_settings->set('context_provider', '@group_context_path_prefix.path_prefix_context:group');
    $group_sites_settings->save();
  }

  /**
   * Tests the this module in integration with a common set of enabled modules.
   */
  public function testIntegration(): void {
    $assert = $this->assertSession();
    $this->drupalLogin($this->testUser);
    // Visit node A via the group A prefix.
    $group_a_path_prefix = GroupPathPrefix::get($this->groupA);
    $node_a_id = $this->nodeA->id();
    $this->drupalGet(Url::fromUri("internal:{$group_a_path_prefix}/node/{$node_a_id}"));
    $assert->pageTextContains($this->nodeA->label());
    // Visit node B via the group B prefix.
    $group_b_path_prefix = GroupPathPrefix::get($this->groupB);
    $node_b_id = $this->nodeB->id();
    $this->drupalGet(Url::fromUri("internal:{$group_b_path_prefix}/node/{$node_b_id}"));
    $assert->pageTextContains($this->nodeB->label());
    // Visit node A, using the group B path prefix.
    $this->drupalGet(Url::fromUri("internal:{$group_b_path_prefix}/node/{$node_a_id}"));
    // Should have been redirected to the group A path prefix.
    $assert->pageTextContains($this->nodeA->label());
    self::assertSame("{$this->baseUrl}{$group_a_path_prefix}/node/{$node_a_id}", $this->getSession()->getCurrentUrl());
    // Visit /{groupA}/admin/content; only node A should appear.
    $this->drupalGet("{$group_a_path_prefix}/admin/content");
    $assert->pageTextContains($this->nodeA->label());
    $assert->pageTextNotContains($this->nodeB->label());
    // Vice-versa.
    $this->drupalGet("{$group_b_path_prefix}/admin/content");
    $assert->pageTextNotContains($this->nodeA->label());
    $assert->pageTextContains($this->nodeB->label());
    // The group canonical page should use its prefix.
    $group_canonical_path = $this->groupA->toUrl();
    $this->drupalGet($group_canonical_path);
    self::assertSame("{$this->baseUrl}{$group_a_path_prefix}/group/{$this->groupA->id()}", $this->getSession()->getCurrentUrl());
    $assert->pageTextContains($this->groupA->label());
    // Set the front page to a group with a prefix.
    $this->config('system.site')->set('page.front', "/group/{$this->groupA->id()}")->save();
    $this->drupalGet(Url::fromRoute('<front>'));
    self::assertSame("{$this->baseUrl}/", $this->getSession()->getCurrentUrl());
  }

}
